/*
 * aconvert.cpp:
 *
 * Read an img file and write an .aff file.
 */

#include "afflib.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <zlib.h>
#include <openssl/md5.h>
#include <openssl/sha.h>

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. [Omitted.]
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */



char *progname = "aconvert";

size_t segsize = 16*1024*1024;	// default seg size --- 16MB
int  opt_compress = 1;
int  opt_fake = 0;
int  fake_segments = 0;			// how many fake segments we need
long long bytes_to_convert = 0;
int  opt_batch = 1;
char *opt_ext = "aff";

void usage()
{
    printf("usage: %s [options] raw1 raw2 raw3...  --- batch convert each raw file\n");
    printf("      -s nnnn  -- set the segsize (default %d)\n",segsize);
    printf("      -x       -- don't compress.\n");
    printf("      -g nnn   -- fake input of nnn gigabytes (for testing)\n");
    printf("      -e ext   -- use 'ext' for the converted files (default %s)\n",opt_ext);
    printf("      -w file  -- convert to 'file' (can only convert one at a time)\n");
    exit(0);
}


int convert(char *infile,char *outfile)
{
    /* if outfile is NULL, figure it out from infile */
    if(outfile==0){
	outfile = (char *)alloca(strlen(infile)+16);
	 strcpy(outfile,infile);
	 char *cc = rindex(outfile,'.'); // to strip off extension
	 if(cc){
	     strcpy(cc+1,opt_ext);
	 }
	 else {
	     strcat(outfile,".");
	     strcat(outfile,opt_ext);
	 }
     }

    struct stat so;
    if(stat(outfile,&so)==0){
	 fprintf(stderr,"%s: file exists. Delete it before converting.\n",outfile);
	 return -1;
     }

     /* Figure out how many bytes to convert if we are not faking it*/
     struct stat si;
     if(stat(infile,&si)){
	 err(2,"stat(%s) ",infile);
     }
     if(!opt_fake){
	 bytes_to_convert = si.st_size;
     } else {
	 bytes_to_convert = (unsigned long long)fake_segments * segsize;
     }


     printf("converting %s to %s...\n",infile,outfile);

     FILE *fin = fopen(infile,"rb");
     if(!fin){
	 err(1,"fopen %s: ",infile);
     }


     /* Create the AFF file */
     AFFILE *af = af_open(outfile,O_CREAT|O_RDWR|O_TRUNC,0666);
     if(!af){
	 err(1,"dbopen %s: ",outfile);
     }

     /* Write out the segment size */
     af_setsize(af,segsize,(unsigned long long)0);		// 

     /* Store the bad block flag and indicate that we won't have any bad blocks.
      * The hashes will be updated later.
      */
     unsigned char md5_buf[16];
     unsigned char sha1_buf[20];

     memset(md5_buf,0,sizeof(md5_buf));
     memset(sha1_buf,0,sizeof(sha1_buf));
     af_put_badflag(af);
     af_put_segv(af,AF_MD5,0,md5_buf,16);
     af_put_segv(af,AF_SHA1,0,sha1_buf,20);

     /* Now we are done with metadata; start the conversion... */

     char *data = (char *)malloc(segsize); // buffer to store the data we read
     char *cdata = (char *)malloc(segsize); // buffer to store the compressed data
     if(!data || !cdata){
	 err(1,"malloc");
     }
     int segnum = 0;
     long long bytes_read=0;
     long long bytes_written=0;
     long csegs_written = 0;
     long segs_written = 0;

     /* Set up the MD5 machinery */
     MD5_CTX md5;
     MD5_Init(&md5);

     SHA_CTX sha;
     SHA1_Init(&sha);

     while(!feof(fin)){
	 printf("\rConverting segment %d  ",segnum);
	 printf("%qd/%qd bytes ",bytes_read,bytes_to_convert);
	 printf("%6.2f%% ...",bytes_read*100.0/bytes_to_convert);
	 fflush(stdout);

	 size_t bytes = 0;

	 /* Read the data from the file to convert */
	 if(opt_fake==0){
	     bytes = fread(data,1,segsize,fin);
	 }
	 else {
	     bytes = segsize;
	     fake_segments--;
	     if(fake_segments==0) bytes = 0;
	     sprintf(data,"fake segment %d\n",segnum);
	 }

	 if(bytes<=0) break;

	 MD5_Update(&md5,data,bytes);
	 SHA1_Update(&sha,data,bytes);
	 bytes_read += bytes;

	 /* Compress the data */
	 uLongf destLen = segsize;
	 int cres = -1;			// compression result; default is that we failed

	 if(opt_compress){		// try to compress
	     cres = compress((Bytef *)cdata,&destLen,(Bytef *)data,bytes);
	 }

	 char vbuf[256];
	 sprintf(vbuf,AF_SEG,segnum);
	 if(cres==0 && destLen < bytes){
	     /* Prepare to write out the compressed segment with compression*/
	     af_put_segv(af,vbuf,AF_SEG_COMPRESSED,cdata,destLen);
	     bytes_written += destLen;
	     csegs_written++;
	 }
	 else {
	     /* Prepare to write out the segment without compression */
	     af_put_segv(af,vbuf,0,data,bytes);
	     bytes_written += bytes;
	     segs_written++;
	 }
	 segnum++;
     }

     printf("\n\nDone!\n\n");

     /* Ideally, go back and update the image size */
     af_update_segq(af,AF_IMAGESIZE,bytes_read);


     /* Finish the hash calculations and write to the db */
     MD5_Final(md5_buf,&md5);
     char buf[256];
     af_update_segv(af,AF_MD5,0,md5_buf,16);
     printf("md5: %s\n",af_hexbuf(buf,md5_buf,16));

     SHA1_Final(sha1_buf,&sha);
     af_update_segv(af,AF_SHA1,0,sha1_buf,20);
     printf("sha1: %s\n",af_hexbuf(buf,sha1_buf,20));
     af_close(af);

     if(stat(outfile,&so)){
	 err(1,"stat(%s) ",outfile);
     }

     printf("bytes read: %qd  bytes written: %qd\n",bytes_read,bytes_written);
     printf("Total segments: %d  (%d compressed)\n",segnum,csegs_written);
     printf("Conversion finished.\n");
     printf("AFF file is %qd bytes. Net compression radio with overhead: %4.1f%%\n",
	    so.st_size,((float)bytes_read-so.st_size)*100.0/bytes_read);
     printf("\n\n");

     /* Set the utime on the resulting file */
     struct timeval times[2];
     memset(&times,0,sizeof(times));
     times[0].tv_sec = si.st_mtime;
     times[1].tv_sec = si.st_mtime;
     utimes(outfile,times);
}


int main(int argc,char **argv)
{
    int bflag, ch;
    char *outfile = 0;
    unsigned char ibuf[256];

    progname = argv[0];
    bflag = 0;
    while ((ch = getopt(argc, argv, "s:xg:h?ew:")) != -1) {
	switch (ch) {
	case 's':
	    segsize = atoi(optarg);
	    break;
	case 'x':
	    opt_compress=0;
	    break;
	case 'g':
	    opt_fake = 1;
	    fake_segments = (1024*1024*1024 / segsize) * atoi(optarg);

	    printf("Will need %d fake segments...\n",fake_segments);
	    break;
	case 'e':
	    opt_ext = optarg;
	    break;
	case 'w':
	    outfile = optarg;
	    break;
	case 'h':
	case '?':
	default:
	    usage();
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<1){
	usage();
    }

    argc--;

    if(outfile){
	convert(*argv,outfile);
	return 0;			// only can convert a single file
    }

    while(*argv){
	convert(*argv++,0);
	argc--;
    }
    return 0;
}
